﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;
using GameUtils.Graphics;
using GameUtils.Graphics.Data;
using SharpDX;
using SharpDX.Direct3D;
using SharpDX.DirectWrite;
using D2D = SharpDX.Direct2D1;
using D3D = SharpDX.Direct3D11;
using DXGI = SharpDX.DXGI;
using Factory = GameUtils.Graphics.Data.Factory;
using Font = GameUtils.Graphics.Font;
using WordWrapping = GameUtils.Graphics.WordWrapping;

namespace GameUtils.Renderers.Direct2D1_1
{
    public class Direct2D1_1Renderer : Renderer
    {
        Control surface;
        D2D.Factory1 d2dFactory;
        D2D.Device d2dDevice;
        D2D.DeviceContext d2dDeviceContext;
        D2D.Bitmap1 d2dSurface;
        SharpDX.DirectWrite.Factory textFactory;
        D3D.Device d3dDevice;
        DXGI.Device1 dxgiDevice;
        DXGI.SwapChain1 swapChain;
        Stack<SharpDX.Direct2D1.Layer> layers;
        bool initialized;
        Direct2DFactory factory;
        bool resized;
        bool forceImmediatelyDraw;
        int syncInterval;

        protected override Factory Factory
        {
            get { return factory ?? (factory = new Direct2DFactory(d2dFactory, d2dDeviceContext, textFactory)); }
        }

        public override Version PlatformVersion
        {
            get { return new Version(6, 1, 7601); }
        }

        public override Math.Rectangle SurfaceBounds
        {
            get { return new Math.Rectangle(0, 0, surface.ClientSize.Width, surface.ClientSize.Height); }
        }
        
        protected override void Initialize(Control surface, PresentationOption presentationOption, AntiAliasMode antiAliasMode)
        {
            this.surface = surface;
            surface.Resize += (sender, e) => resized = true;
            layers = new Stack<SharpDX.Direct2D1.Layer>();
            syncInterval = presentationOption == PresentationOption.Limited ? 1 : 0;

            d3dDevice = new D3D.Device(DriverType.Hardware, D3D.DeviceCreationFlags.BgraSupport);
            dxgiDevice = d3dDevice.QueryInterface<DXGI.Device1>();
            var desc = new DXGI.SwapChainDescription1()
            {
                Width = 0,
                Height = 0,
                Format = DXGI.Format.B8G8R8A8_UNorm,
                Stereo = false,
                SampleDescription = new DXGI.SampleDescription(1, 0),
                Usage = DXGI.Usage.RenderTargetOutput,
                BufferCount = 2,
                Scaling = DXGI.Scaling.Stretch,
                SwapEffect = DXGI.SwapEffect.Discard,
                Flags = DXGI.SwapChainFlags.None
            };
            DXGI.Factory2 dxgiFactory = dxgiDevice.Adapter.GetParent<DXGI.Factory2>();
            swapChain = dxgiFactory.CreateSwapChainForHwnd(d3dDevice, surface.Handle, ref desc, null, null);
            dxgiDevice.MaximumFrameLatency = 1;

            d2dFactory = new D2D.Factory1(D2D.FactoryType.MultiThreaded);
            d2dDevice = new D2D.Device(d2dFactory, dxgiDevice);
            d2dDeviceContext = new D2D.DeviceContext(d2dDevice, D2D.DeviceContextOptions.EnableMultithreadedOptimizations);
            d2dDeviceContext.AntialiasMode = antiAliasMode == AntiAliasMode.AntiAlias ? D2D.AntialiasMode.PerPrimitive : D2D.AntialiasMode.Aliased;
            d2dDeviceContext.TextAntialiasMode = antiAliasMode == AntiAliasMode.AntiAlias ? D2D.TextAntialiasMode.Cleartype : D2D.TextAntialiasMode.Aliased;
            d2dDeviceContext.DotsPerInch = new Size2F(96, 96);
            var props = new D2D.BitmapProperties1(new D2D.PixelFormat(DXGI.Format.B8G8R8A8_UNorm, D2D.AlphaMode.Ignore),
                96, 96, D2D.BitmapOptions.Target | D2D.BitmapOptions.CannotDraw);
            DXGI.Surface1 dxgiSurface = swapChain.GetBackBuffer<DXGI.Surface1>(0);
            d2dSurface = new D2D.Bitmap1(d2dDeviceContext, dxgiSurface, props);
            dxgiSurface.Dispose();
            d2dDeviceContext.Target = d2dSurface;
            textFactory = new SharpDX.DirectWrite.Factory();

            initialized = true;
        }

        protected override void BeginRender()
        {
            if (!initialized)
                throw new InvalidOperationException("Der Renderer muss initialisiert werden, bevor er verwendet werden kann.");

            d2dDeviceContext.BeginDraw();
        }

        protected override void EndRender()
        {
            d2dDeviceContext.EndDraw();

            if (resized)
            {
                swapChain.Present(0, DXGI.PresentFlags.None);

                d2dDeviceContext.Target = null;
                d2dSurface.Dispose();
                swapChain.ResizeBuffers(0, 0, 0, DXGI.Format.Unknown, DXGI.SwapChainFlags.None);
                DXGI.Surface1 dxgiSurface = swapChain.GetBackBuffer<DXGI.Surface1>(0);
                var props = new D2D.BitmapProperties1(new D2D.PixelFormat(DXGI.Format.B8G8R8A8_UNorm, D2D.AlphaMode.Ignore),
                96, 96, D2D.BitmapOptions.Target | D2D.BitmapOptions.CannotDraw);
                d2dSurface = new D2D.Bitmap1(d2dDeviceContext, dxgiSurface, props);
                dxgiSurface.Dispose();
                d2dDeviceContext.Target = d2dSurface;
                resized = false;
                forceImmediatelyDraw = true;
            }
            else if (forceImmediatelyDraw)
            {
                swapChain.Present(0, DXGI.PresentFlags.None);
                forceImmediatelyDraw = false;
            }
            else
            {
                swapChain.Present(syncInterval, DXGI.PresentFlags.None);
            }
        }

        private D2D.Brush GetD2DBrush<T>(Brush<T> brush) where T : BrushData
        {
            if (brush == null)
                throw new ArgumentNullException("brush");

            try
            {
                return ((IDirect2DBrush)this.GetResourceData(brush)).Brush;
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Brush-Objekt ist nicht kompatibel.", ex);
            }
        }

        private D2D.StrokeStyle1 GetD2DStrokeStyle(StrokeStyle? strokeStyle)
        {
            if (!strokeStyle.HasValue)
                return null;

            var props = new D2D.StrokeStyleProperties1();
            props.StartCap = (D2D.CapStyle)strokeStyle.Value.StartCap;
            props.EndCap = (D2D.CapStyle)strokeStyle.Value.EndCap;
            props.DashCap = (D2D.CapStyle)strokeStyle.Value.DashCap;
            props.DashOffset = strokeStyle.Value.DashOffset;
            props.MiterLimit = strokeStyle.Value.MiterLimit;
            props.LineJoin = (D2D.LineJoin)strokeStyle.Value.LineJoin;
            switch (strokeStyle.Value.DashStyle)
            {
                case DashStyle.Custom:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.Custom;
                    break;
                case DashStyle.Dash:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.Dash;
                    break;
                case DashStyle.DashDot:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.DashDot;
                    break;
                case DashStyle.DashDotDot:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.DashDotDot;
                    break;
                case DashStyle.Dot:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.Dot;
                    break;
                default:
                    props.DashStyle = SharpDX.Direct2D1.DashStyle.Solid;
                    break;
            }

            if (strokeStyle.Value.DashStyle == DashStyle.Custom)
                return new D2D.StrokeStyle1(d2dFactory, props, strokeStyle.Value.DashPattern);
            else
                return new D2D.StrokeStyle1(d2dFactory, props);
        }

        private void DrawLineInner<T>(Math.Vector2 p1, Math.Vector2 p2, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            D2D.Brush d2dBrush = GetD2DBrush(brush);
            using (D2D.StrokeStyle1 d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                d2dDeviceContext.DrawLine(Direct2DHelper.Vector2ToPointF(p1), Direct2DHelper.Vector2ToPointF(p2), d2dBrush, strokeWidth, d2dStrokeStyle);
            }
        }

        public override void DrawLine<T>(Math.Vector2 p1, Math.Vector2 p2, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawLineInner(p1, p2, brush, strokeWidth, null);
        }

        public override void DrawLine<T>(Math.Vector2 p1, Math.Vector2 p2, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawLineInner(p1, p2, brush, strokeWidth, strokeStyle);
        }

        private void DrawEllipseInner<T>(Math.Ellipse ellipse, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            D2D.Brush d2dBrush = GetD2DBrush(brush);
            using (D2D.StrokeStyle1 d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                d2dDeviceContext.DrawEllipse(Direct2DHelper.GraphicsEllipseToD2DEllipse(ellipse), d2dBrush, strokeWidth, d2dStrokeStyle);
            }
        }

        public override void DrawEllipse<T>(Math.Ellipse ellipse, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawEllipseInner(ellipse, brush, strokeWidth, null);
        }

        public override void DrawEllipse<T>(Math.Ellipse ellipse, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawEllipseInner(ellipse, brush, strokeWidth, strokeStyle);
        }

        public override void FillEllipse<T>(Math.Ellipse ellipse, Brush<T> brush)
        {
            D2D.Brush d2dBrush = GetD2DBrush(brush);
            d2dDeviceContext.FillEllipse(Direct2DHelper.GraphicsEllipseToD2DEllipse(ellipse), d2dBrush);
        }

        private void DrawRectangleInner<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            D2D.Brush d2dBrush = GetD2DBrush(brush);
            using (D2D.StrokeStyle1 d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                d2dDeviceContext.DrawRectangle(Direct2DHelper.ConvertRectangle(rect), d2dBrush, strokeWidth, d2dStrokeStyle);
            }
        }

        public override void DrawRectangle<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawRectangleInner(rect, brush, strokeWidth, null);
        }

        public override void DrawRectangle<T>(Math.Rectangle rect, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawRectangleInner(rect, brush, strokeWidth, strokeStyle);
        }

        public override void FillRectangle<T>(Math.Rectangle rect, Brush<T> brush)
        {
            D2D.Brush d2dBrush = GetD2DBrush(brush);
            d2dDeviceContext.FillRectangle(Direct2DHelper.ConvertRectangle(rect), d2dBrush);
        }

        private void DrawPolygonInner<T>(Math.Polygon polygon, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            if (!polygon.IsValid)
                throw new ArgumentException("polygon", "Das übergebene Polygon ist nicht gültig.");

            D2D.Brush d2dBrush = GetD2DBrush(brush);
            using (D2D.StrokeStyle1 d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                using (var geometry = new D2D.PathGeometry(d2dFactory))
                {
                    using (var sink = geometry.Open())
                    {
                        sink.BeginFigure(Direct2DHelper.Vector2ToPointF(polygon.Points[0]), D2D.FigureBegin.Hollow);

                        for (int i = 1; i < polygon.Points.Length; i++)
                            sink.AddLine(Direct2DHelper.Vector2ToPointF(polygon.Points[i]));

                        sink.EndFigure(SharpDX.Direct2D1.FigureEnd.Closed);
                        sink.Close();

                        d2dDeviceContext.DrawGeometry(geometry, d2dBrush, strokeWidth, d2dStrokeStyle);
                    }
                }
            }
        }

        public override void DrawPolygon<T>(Math.Polygon polygon, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawPolygonInner(polygon, brush, strokeWidth, null);
        }

        public override void DrawPolygon<T>(Math.Polygon polygon, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawPolygonInner(polygon, brush, strokeWidth, strokeStyle);
        }

        public override void FillPolygon<T>(Math.Polygon polygon, Brush<T> brush)
        {
            if (!polygon.IsValid)
                throw new ArgumentException("polygon", "Das übergebene Polygon ist nicht gültig.");

            D2D.Brush d2dBrush = GetD2DBrush(brush);
            using (var geometry = new D2D.PathGeometry1(d2dFactory))
            {
                using (var sink = geometry.Open())
                {
                    sink.BeginFigure(Direct2DHelper.Vector2ToPointF(polygon.Points[0]), D2D.FigureBegin.Filled);

                    for (int i = 1; i < polygon.Points.Length; i++)
                        sink.AddLine(Direct2DHelper.Vector2ToPointF(polygon.Points[i]));

                    sink.EndFigure(SharpDX.Direct2D1.FigureEnd.Closed);
                    sink.Close();

                    d2dDeviceContext.FillGeometry(geometry, d2dBrush);
                }
            }
        }

        private void DrawTextInner<T>(string text, Font font, Brush<T> brush, Math.Rectangle rect, Graphics.TextFormat format) where T : BrushData
        {
            D2D.Brush d2dBrush = GetD2DBrush(brush);

            if (font == null)
                throw new ArgumentNullException("font");
            Direct2DFont d2dFont;
            try
            {
                d2dFont = (Direct2DFont)this.GetResourceData(font);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Font-Objekt ist nicht kompatibel.", ex);
            }
                

            SharpDX.DirectWrite.TextFormat d2dFormat = d2dFont.GetTextFormat();
            if (format != null)
            {
                switch (format.HorizontalAlignment)
                {
                    case Graphics.HorizontalAlignment.Leading:
                        d2dFormat.TextAlignment = TextAlignment.Leading;
                        break;
                    case Graphics.HorizontalAlignment.Center:
                        d2dFormat.TextAlignment = TextAlignment.Center;
                        break;
                    case Graphics.HorizontalAlignment.Trailing:
                        d2dFormat.TextAlignment = TextAlignment.Trailing;
                        break;
                }
                switch (format.VerticalAlignment)
                {
                    case VerticalAlignment.Top:
                        d2dFormat.ParagraphAlignment = ParagraphAlignment.Near;
                        break;
                    case VerticalAlignment.Center:
                        d2dFormat.ParagraphAlignment = ParagraphAlignment.Center;
                        break;
                    case VerticalAlignment.Bottom:
                        d2dFormat.ParagraphAlignment = ParagraphAlignment.Far;
                        break;
                }
                d2dFormat.WordWrapping = (format.WordWrapping == WordWrapping.Wrap) ? SharpDX.DirectWrite.WordWrapping.Wrap : SharpDX.DirectWrite.WordWrapping.NoWrap;
            }
            else
            {
                d2dFormat.TextAlignment = TextAlignment.Leading;
                d2dFormat.ParagraphAlignment = ParagraphAlignment.Near;
                d2dFormat.WordWrapping = SharpDX.DirectWrite.WordWrapping.Wrap;
            }

            D2D.DrawTextOptions options = format == null || format.WordWrapping != WordWrapping.Clip ? D2D.DrawTextOptions.NoSnap : D2D.DrawTextOptions.Clip | D2D.DrawTextOptions.NoSnap;
            d2dDeviceContext.DrawText(text, d2dFormat, Direct2DHelper.ConvertRectangle(rect), d2dBrush, options);
        }

        public override void DrawText<T>(string text, Font font, Brush<T> brush, Math.Rectangle rect, Graphics.TextFormat format = null)
        {
            this.DrawTextInner(text, font, brush, rect, format);
        }

        public override void DrawText<T>(string text, Font font, Brush<T> brush, Math.Vector2 pt)
        {
            this.DrawTextInner(text, font, brush, new Math.Rectangle(pt.X, pt.Y, float.MaxValue, float.MaxValue), null);
        }

        private void DrawTextureInner(Texture texture, Math.Rectangle destRect, Math.Rectangle? sourceRect, float opacity, InterpolationMode interpolationMode)
        {
            if (texture == null)
                throw new ArgumentNullException("texture");
            Direct2DTexture d2dTexture;
            try
            {
                d2dTexture = (Direct2DTexture)this.GetResourceData(texture);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Diese Texture ist nicht kompatibel.", ex);
            }
                
            D2D.Bitmap1 d2dBitmap = d2dTexture.GetBitmap();

            d2dDeviceContext.DrawBitmap(d2dBitmap, Direct2DHelper.ConvertRectangle(destRect), opacity, Direct2DHelper.ConvertInterpolationMode(interpolationMode),
                sourceRect.HasValue ? Direct2DHelper.ConvertRectangle(sourceRect.Value) : (RectangleF?)null, null);
        }

        public override void DrawTexture(Texture texture, Math.Rectangle destRect, Math.Rectangle? sourceRect = null, float opacity = 1.0f, InterpolationMode interpolationMode = InterpolationMode.HighQualityCubic)
        {
            this.DrawTextureInner(texture, destRect, sourceRect, opacity, interpolationMode);
        }

        public override void DrawTexture(Texture texture, Math.Vector2 pt, float opacity = 1.0f)
        {
            this.DrawTextureInner(texture, new Math.Rectangle(pt.X, pt.Y, texture.Width, texture.Height), null, opacity, InterpolationMode.HighQualityCubic);
        }

        private void DrawGeometryInner<T>(Geometry geometry, Brush<T> brush, float strokeWidth, StrokeStyle? strokeStyle) where T : BrushData
        {
            D2D.Brush d2dBrush = GetD2DBrush(brush);

            if (geometry == null)
                throw new ArgumentNullException("geometry");
            Direct2DGeometry d2dGeometry;
            try
            {
                d2dGeometry = (Direct2DGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            using (D2D.StrokeStyle1 d2dStrokeStyle = this.GetD2DStrokeStyle(strokeStyle))
            {
                d2dDeviceContext.DrawGeometry(d2dGeometry.Geometry, d2dBrush, strokeWidth, d2dStrokeStyle);
            }
        }

        public override void DrawGeometry<T>(Geometry geometry, Brush<T> brush, float strokeWidth = 1.0f)
        {
            this.DrawGeometryInner(geometry, brush, strokeWidth, null);
        }

        public override void DrawGeometry<T>(Geometry geometry, Brush<T> brush, float strokeWidth, StrokeStyle strokeStyle)
        {
            this.DrawGeometryInner(geometry, brush, strokeWidth, strokeStyle);
        }

        public override void FillGeometry<T>(Geometry geometry, Brush<T> brush)
        {
            D2D.Brush d2dBrush = GetD2DBrush(brush);

            if (geometry == null)
                throw new ArgumentNullException("geometry");
            Direct2DGeometry d2dGeometry;
            try
            {
                d2dGeometry = (Direct2DGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            d2dDeviceContext.FillGeometry(d2dGeometry.Geometry, d2dBrush);
        }

        public override void Clear(Graphics.Color4 color)
        {
            d2dDeviceContext.Clear(Direct2DHelper.GraphicsColorToD2DColor(color));
        }

        public override void PushClip(Geometry geometry)
        {
            if (geometry == null)
                throw new ArgumentNullException("geometry");
            Direct2DGeometry d2dGeometry;
            try
            {
                d2dGeometry = (Direct2DGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            var layer = new SharpDX.Direct2D1.Layer(d2dDeviceContext);
            layers.Push(layer);
            var layerParams = new D2D.LayerParameters();
            layerParams.GeometricMask = d2dGeometry.Geometry;
            layerParams.MaskAntialiasMode = D2D.AntialiasMode.PerPrimitive;
            layerParams.LayerOptions = D2D.LayerOptions.None;
            layerParams.Opacity = 1;
            d2dDeviceContext.PushLayer(ref layerParams, layer);
        }

        public override void PushClip(Math.Rectangle rect)
        {
            var layer = new SharpDX.Direct2D1.Layer(d2dDeviceContext);
            layers.Push(layer);
            var layerParams = new D2D.LayerParameters();
            layerParams.ContentBounds = Direct2DHelper.ConvertRectangle(rect);
            layerParams.LayerOptions = D2D.LayerOptions.None;
            layerParams.Opacity = 1;
            d2dDeviceContext.PushLayer(ref layerParams, layer);
        }

        public override void PopClip()
        {
            d2dDeviceContext.PopLayer();
            layers.Pop().Dispose();
        }

        public override void SetTransform(Math.Matrix2x3 matrix)
        {
            var d2dMatrix = new Matrix3x2();
            d2dMatrix.TranslationVector = new Vector2(matrix.OffsetX, matrix.OffsetY);
            d2dMatrix.ScaleVector = new Vector2(matrix[0, 0], matrix[1, 1]);
            d2dMatrix.M12 = matrix[1, 0];
            d2dMatrix.M21 = matrix[0, 1];
            d2dDeviceContext.Transform = d2dMatrix;
        }

        public override void ResetTransform()
        {
            d2dDeviceContext.Transform = Matrix3x2.Identity;
        }

        public override Math.Vector2 MeasureText(string text, Font font)
        {
            if (font == null)
                throw new ArgumentNullException("font");
            Direct2DFont d2dFont;
            try
            {
                d2dFont = (Direct2DFont)this.GetResourceData(font);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Font-Objekt ist nicht kompatibel.", ex);
            }
            SharpDX.DirectWrite.TextFormat d2dFormat = d2dFont.GetTextFormat();
            d2dFormat.ParagraphAlignment = ParagraphAlignment.Near;
            d2dFormat.TextAlignment = TextAlignment.Leading;
            d2dFormat.WordWrapping = SharpDX.DirectWrite.WordWrapping.NoWrap;

            using (var layout = new TextLayout(textFactory, text, d2dFormat, float.MaxValue, float.MaxValue))
            {
                return new Math.Vector2(layout.Metrics.Width, layout.Metrics.Height);
            }
        }

        protected override void Dispose(bool disposing)
        {
            while (layers != null && layers.Count > 0)
                layers.Pop().Dispose();
            if (disposing)
            {
                factory = null;
                layers = null;
            }
            if (textFactory != null) textFactory.Dispose();
            if (d2dSurface != null) d2dSurface.Dispose();
            if (d2dDeviceContext != null) d2dDeviceContext.Dispose();
            if (d2dFactory != null) d2dFactory.Dispose();
            if (d3dDevice != null) d3dDevice.Dispose();
        }
    }
}
