﻿using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using GameUtils.Graphics;
using GameUtils.Graphics.Data;
using GameUtils.Math;

namespace GameUtils.Renderers.Gdi
{
    internal class GdiGeometry : GeometryData
    {
        readonly Region region;
        readonly GraphicsPath path;
        GraphicsPath figurePath;
        PointF lastPoint;

        public GdiGeometry()
        {
            path = new GraphicsPath();
            region = new Region();
            region.MakeEmpty();
        }

        internal GraphicsPath Path
        {
            get { return path; }
        }

        internal Region Region
        {
            get { return region; }
        }

        protected override void AddArc(Arc arc)
        {
            throw new NotImplementedException();
        }

        protected override void AddLine(Vector2 point)
        {
            PointF drawingPoint = new PointF(point.X, point.Y);
            figurePath.AddLine(lastPoint, drawingPoint);
            lastPoint = drawingPoint;
        }

        protected override void AddLines(Vector2[] points)
        {
            PointF[] drawingPoints = new PointF[points.Length + 1];
            drawingPoints[0] = lastPoint;
            for (int i = 1; i <= points.Length; i++)
                drawingPoints[i] = new PointF(points[i - 1].X, points[i - 1].Y);
            figurePath.AddLines(drawingPoints);
            lastPoint = drawingPoints[drawingPoints.Length - 1];
        }

        protected override void BeginFigure(Vector2 point, FigureBegin begin)
        {
            figurePath = new GraphicsPath(FillMode.Winding);
            lastPoint = new PointF(point.X, point.Y);
        }

        protected override void EndFigure(FigureEnd end)
        {
            if (end == FigureEnd.Closed)
                figurePath.CloseFigure();
            else
                figurePath.StartFigure();
            path.AddPath(figurePath, false);
            region.Union(figurePath);
            figurePath.Dispose();
        }

        protected override void AddBezier(Bezier bezier)
        {
            PointF endPoint = new PointF(bezier.EndPoint.X, bezier.EndPoint.Y);
            figurePath.AddBezier(lastPoint, new PointF(bezier.ControlPoint1.X, bezier.ControlPoint1.Y), new PointF(bezier.ControlPoint2.X, bezier.ControlPoint2.Y), endPoint);
            lastPoint = endPoint;
        }

        protected override void AddBeziers(Bezier[] beziers)
        {
            PointF[] drawingPoints = new PointF[beziers.Length * 3 + 1];
            drawingPoints[0] = lastPoint;
            for (int i = 0, j = 1; i < beziers.Length; i++, j += 3)
            {
                drawingPoints[j] = new PointF(beziers[i].ControlPoint1.X, beziers[i].ControlPoint1.Y);
                drawingPoints[j + 1] = new PointF(beziers[i].ControlPoint2.X, beziers[i].ControlPoint2.Y);
                drawingPoints[j + 2] = new PointF(beziers[i].EndPoint.X, beziers[i].EndPoint.Y);
            }
            figurePath.AddBeziers(drawingPoints);
            lastPoint = drawingPoints[drawingPoints.Length - 1];
        }

        protected override void AddCurve(Vector2[] points, float tension)
        {
            PointF[] drawingPoints = new PointF[points.Length + 1];
            drawingPoints[0] = lastPoint;
            for (int i = 0; i < points.Length; i++)
                drawingPoints[i + 1] = new PointF(points[i].X, points[i].Y);
            figurePath.AddCurve(drawingPoints, (1 - tension) / 2);
            lastPoint = drawingPoints[drawingPoints.Length - 1];
        }

        protected override void AddGeometry(Geometry geometry, GeometryCombineMode combineMode)
        {
            GdiGeometry other;
            try
            {
                other = (GdiGeometry)this.GetResourceData(geometry);
            }
            catch (InvalidCastException ex)
            {
                throw new ArgumentException("Dieses Geometry-Objekt ist nicht kompatibel.", ex);
            }

            switch (combineMode)
            {
                case GeometryCombineMode.Intersect:
                    region.Intersect(other.region);
                    break;
                case GeometryCombineMode.Xor:
                    region.Xor(other.region);
                    break;
                case GeometryCombineMode.Exclude:
                    region.Exclude(other.region);
                    break;
                default:
                    region.Union(other.region);
                    break;
            }
            path.AddPath(other.path, false);
        }

        protected override void AddRectangle(Math.Rectangle rect)
        {
            var gdiRect = new RectangleF(rect.X, rect.Y, rect.Width, rect.Height);
            region.Union(gdiRect);
            path.AddRectangle(gdiRect);
        }

        protected override void AddEllipse(Ellipse ellipse)
        {
            var rect = new RectangleF(ellipse.Center.X - ellipse.RadiusX, ellipse.Center.Y - ellipse.RadiusY, ellipse.RadiusX * 2, ellipse.RadiusY * 2);
            path.AddEllipse(rect);
            var ellipsePath = new GraphicsPath();
            ellipsePath.AddEllipse(rect);
            region.Union(ellipsePath);
            ellipsePath.Dispose();
        }

        protected override void Transform(Matrix2x3 matrix)
        {
            var transformationMatrix = new Matrix(matrix[0, 0], matrix[1, 0], matrix[0, 1], matrix[1, 1], matrix.OffsetX, matrix.OffsetY);
            path.Transform(transformationMatrix);
            region.Transform(transformationMatrix);
            transformationMatrix.Dispose();
        }

        protected override void Close()
        { }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                region.Dispose();
                path.Dispose();
            }
        }
    }
}
