﻿using System;
using GameUtils.Graphics.Data;
using GameUtils.Math;

namespace GameUtils.Graphics
{
    public class RadialGradientBrush : Brush<RadialGradientBrushData>
    {
        GradientStop[] gradientStops;
        Vector2 center;
        float radiusX;
        float radiusY;
        Vector2 originOffset;
        bool stopsChanged;
        bool centerChanged;
        bool radiusXChanged;
        bool radiusYChanged;
        bool originOffsetChanged;
        readonly object locker;

        /// <summary>
        /// A list of gradient stops defining the colors with their positions in the gradient.
        /// </summary>
        public GradientStop[] GradientStops
        {
            get { return gradientStops; }
            set
            {
                if (value == null)
                    throw new ArgumentNullException("value");
                if (value.Length < 2)
                    throw new ArgumentException("Es müssen mindestens zwei Verlaufsfarben angegeben sein.", "value");

                lock (locker)
                {
                    gradientStops = value;
                    stopsChanged = true;
                }
            }
        }

        /// <summary>
        /// The ellipse's center.
        /// </summary>
        public Vector2 Center
        {
            get { return center; }
            set
            {
                lock (locker)
                {
                    center = value;
                    centerChanged = true;
                }
            }
        }

        /// <summary>
        /// The ellipse's horizontal radius.
        /// </summary>
        public float RadiusX
        {
            get { return radiusX; }
            set
            {
                lock (locker)
                {
                    radiusX = value;
                    radiusXChanged = true;
                }
            }
        }

        /// <summary>
        /// The ellipse's vertical radius.
        /// </summary>
        public float RadiusY
        {
            get { return radiusY; }
            set
            {
                lock (locker)
                {
                    radiusY = value;
                    radiusYChanged = true;
                }
            }
        }
        
        /// <summary>
        /// The gradient's center relative to the ellipse's center.
        /// </summary>
        public Vector2 OriginOffset
        {
            get { return originOffset; }
            set
            {
                lock (locker)
                {
                    originOffset = value;
                    originOffsetChanged = true;
                }
            }
        }

        /// <summary>
        /// Creates a new radial gradient brush.
        /// </summary>
        public RadialGradientBrush(GradientStop[] stops, Vector2 center, float radiusX, float radiusY)
        {
            if (stops == null)
                throw new ArgumentNullException("stops");
            if (stops.Length < 2)
                throw new ArgumentException("Es müssen mindestens zwei Verlaufsfarben angegeben sein.", "stops");

            gradientStops = stops;
            this.center = center;
            this.radiusX = radiusX;
            this.radiusY = radiusY;
            locker = new object();
        }

        protected override void ApplyChanges()
        {
            lock (locker)
            {
                if (stopsChanged)
                {
                    Data.GradientStops = gradientStops;
                    stopsChanged = false;
                }
                if (centerChanged)
                {
                    Data.Center = center;
                    centerChanged = false;
                }
                if (radiusXChanged)
                {
                    Data.RadiusX = radiusX;
                    radiusXChanged = false;
                }
                if (radiusYChanged)
                {
                    Data.RadiusY = radiusY;
                    radiusYChanged = false;
                }
                if (originOffsetChanged)
                {
                    Data.OriginOffset = originOffset;
                    originOffsetChanged = false;
                }
            }

            base.ApplyChanges();
        }

        protected override RadialGradientBrushData CreateData(Factory factory)
        {
            lock (locker)
            {
                stopsChanged = false;
                centerChanged = false;
                radiusXChanged = false;
                radiusYChanged = false;
                originOffsetChanged = false;
                RadialGradientBrushData data = factory.CreateRadialGradientBrush(gradientStops, center, radiusX, radiusY);
                data.OriginOffset = originOffset;
                return data;
            }
        }
    }
}
